// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test5pull.cpp
// Purpose:	Time Measurements for Rule Implementations
// Last Change:	06.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 5:
//
// p1(A) :- q(A).
// p2(A,1) :- p1(A).
// p2(A,2) :- p1(A).
// p3(A,B,1) :- p2(A,B).
// p3(A,B,2) :- p2(A,B).
// p4(A,B,C,1) :- p3(A,B,C).
// p4(A,B,C,2) :- p3(A,B,C).
// p5(A,B,C,D,1) :- p4(A,B,C,D).
// p5(A,B,C,D,2) :- p4(A,B,C,D).
// p6(A,B,C,D,E,1) :- p5(A,B,C,D,E).
// p6(A,B,C,D,E,2) :- p5(A,B,C,D,E).
// p7(A,B,C,D,E,F,1) :- p6(A,B,C,D,E,F).
// p7(A,B,C,D,E,F,2) :- p6(A,B,C,D,E,F).

// q := {(i)|i = 1,...,50000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 50000;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() const { return i; }

	void close() {}

private:
	int i;
};


// ----------------------------------------------------------------------------
// Predicate p1:
// ----------------------------------------------------------------------------
//
// p1(A) :- q(A).

class p1_cursor {
public:
	p1_cursor()
	{
	}

	void open()
	{
		c.open();
	}

	bool fetch()
	{
		return c.fetch();
	}

	int col_1() const { return c.col_1(); }

	void close()
	{
		c.close();
	}

private:
	q_cursor c;
};


// ----------------------------------------------------------------------------
// Predicate p2:
// ----------------------------------------------------------------------------
//
// p2(A,1) :- p1(A).
// p2(A,2) :- p1(A).

class p2_cursor {
public:
	p2_cursor()
	{
	}

	void open()
	{
		c.open();
		rule = 1;
	}

	bool fetch()
	{
		if(c.fetch())
			return true;
		else if(rule == 1) {
			c.close();
			c.open();
			rule = 2;
			return c.fetch();
		}
		else
			return false;
	}

	int col_1() const { return c.col_1(); }
	int col_2() const { return rule; }

	void close()
	{
		c.close();
	}

private:
	p1_cursor c;
	int rule;
};


// ----------------------------------------------------------------------------
// Predicate p3:
// ----------------------------------------------------------------------------
//
// p3(A,B,1) :- p2(A,B).
// p3(A,B,2) :- p2(A,B).

class p3_cursor {
public:
	p3_cursor()
	{
	}

	void open()
	{
		c.open();
		rule = 1;
	}

	bool fetch()
	{
		if(c.fetch())
			return true;
		else if(rule == 1) {
			c.close();
			c.open();
			rule = 2;
			return c.fetch();
		}
		else
			return false;
	}

	int col_1() const { return c.col_1(); }
	int col_2() const { return c.col_2(); }
	int col_3() const { return rule; }

	void close()
	{
		c.close();
	}

private:
	p2_cursor c;
	int rule;
};


// ----------------------------------------------------------------------------
// Predicate p4:
// ----------------------------------------------------------------------------
//
// p4(A,B,C,1) :- p3(A,B,C).
// p4(A,B,C,2) :- p3(A,B,C).

class p4_cursor {
public:
	p4_cursor()
	{
	}

	void open()
	{
		c.open();
		rule = 1;
	}

	bool fetch()
	{
		if(c.fetch())
			return true;
		else if(rule == 1) {
			c.close();
			c.open();
			rule = 2;
			return c.fetch();
		}
		else
			return false;
	}

	int col_1() const { return c.col_1(); }
	int col_2() const { return c.col_2(); }
	int col_3() const { return c.col_3(); }
	int col_4() const { return rule; }

	void close()
	{
		c.close();
	}

private:
	p3_cursor c;
	int rule;
};


// ----------------------------------------------------------------------------
// Predicate p5:
// ----------------------------------------------------------------------------
//
// p5(A,B,C,D,1) :- p4(A,B,C,D).
// p5(A,B,C,D,2) :- p4(A,B,C,D).

class p5_cursor {
public:
	p5_cursor()
	{
	}

	void open()
	{
		c.open();
		rule = 1;
	}

	bool fetch()
	{
		if(c.fetch())
			return true;
		else if(rule == 1) {
			c.close();
			c.open();
			rule = 2;
			return c.fetch();
		}
		else
			return false;
	}

	int col_1() const { return c.col_1(); }
	int col_2() const { return c.col_2(); }
	int col_3() const { return c.col_3(); }
	int col_4() const { return c.col_4(); }
	int col_5() const { return rule; }

	void close()
	{
		c.close();
	}

private:
	p4_cursor c;
	int rule;
};

// ----------------------------------------------------------------------------
// Predicate p6:
// ----------------------------------------------------------------------------
//
// p6(A,B,C,D,E,1) :- p5(A,B,C,D,E).
// p6(A,B,C,D,E,2) :- p5(A,B,C,D,E).

class p6_cursor {
public:
	p6_cursor()
	{
	}

	void open()
	{
		c.open();
		rule = 1;
	}

	bool fetch()
	{
		if(c.fetch())
			return true;
		else if(rule == 1) {
			c.close();
			c.open();
			rule = 2;
			return c.fetch();
		}
		else
			return false;
	}

	int col_1() const { return c.col_1(); }
	int col_2() const { return c.col_2(); }
	int col_3() const { return c.col_3(); }
	int col_4() const { return c.col_4(); }
	int col_5() const { return c.col_5(); }
	int col_6() const { return rule; }

	void close()
	{
		c.close();
	}

private:
	p5_cursor c;
	int rule;
};

// ----------------------------------------------------------------------------
// Predicate p7:
// ----------------------------------------------------------------------------
//
// p7(A,B,C,D,E,F,1) :- p6(A,B,C,D,E,F).
// p7(A,B,C,D,E,F,2) :- p6(A,B,C,D,E,F).

class p7_cursor {
public:
	p7_cursor()
	{
	}

	void open()
	{
		c.open();
		rule = 1;
	}

	bool fetch()
	{
		if(c.fetch())
			return true;
		else if(rule == 1) {
			c.close();
			c.open();
			rule = 2;
			return c.fetch();
		}
		else
			return false;
	}

	int col_1() const { return c.col_1(); }
	int col_2() const { return c.col_2(); }
	int col_3() const { return c.col_3(); }
	int col_4() const { return c.col_4(); }
	int col_5() const { return c.col_5(); }
	int col_6() const { return c.col_6(); }
	int col_7() const { return rule; }

	void close()
	{
		c.close();
	}

private:
	p6_cursor c;
	int rule;
};

//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1_must_be = 1;
	int c1 = 0;
	int c2 = 0;
	int c3 = 0;
	int c4 = 0;
	int c5 = 0;
	int c6 = 0;
	int c7 = 0;

	cout << "Test of Method 2: Pull [Program 5]\n";
	t.start();
	p7_cursor c;
	c.open();
	int num_tuples = 0;
	while(c.fetch()) {
		c1 = c.col_1();
		if(c1 != c1_must_be) {
			cout << "Error: Wrong value for c1:" << c.col_1()
				<< ", must be: " << c1_must_be << "!\n";
			exit(1);
		}
		if(c1_must_be == Q_TUPLES) {
			c1_must_be = 1;
		}
		else
			c1_must_be++;
		c2 += c.col_2();
		c3 += c.col_3();
		c4 += c.col_4();
		c5 += c.col_5();
		c6 += c.col_6();
		c7 += c.col_7();
		num_tuples++;
	}
	c.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p6:\n\n";
		c.open();
		while(c.fetch()) {
			cout << "p7(" << c.col_1() << ", " << c.col_2()
				<< ", " << c.col_3() << ", " << c.col_4()
				<< ", " << c.col_5() << ", " << c.col_6()
				<< ", " << c.col_7() << ").\n";
		}
		c.close();
		cout << "\n";
	}

	if(c2 != Q_TUPLES * 96 )
		cout << "Error: Wrong column sum c2: " << c2 << "\n";
	if(c3 != c2 || c4 != c2 || c5 != c2 || c6 != c2)
		cout << "Error: Wrong column sum c3, c4, c5, c6!\n";
	cout << "Total number of tuples: " << num_tuples << "\n";
	if(num_tuples != Q_TUPLES * 64)
		cout << "Error: Wrong number of tuples!\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


